unit C64;

interface

uses
 windows,resource,sysutils,
 VIC,CIA;

{$ifdef LIST  -- this is for debug purpose do not use this}
Type
 TListRecord=packed record
  cycles:word;
  TimerA1:word; TIA1:word;
  TimerB1:word; TIB1:word;
  TimerA2:word; TIA2:word;
  TimerB2:word; TIB2:word;
  raster:word;
  PC:word;
  OP,A,X,Y,S,P:byte;
 end;
{$endif}
 {$ifdef debug}
var
  dump:boolean;
 {$endif}
 {$ifdef list}
var
  lst:textfile;
  dos:file of TListRecord;
 {$endif}

Type
 TFlags=set of (Cflag,ZFlag,IFlag,DFlag,BFlag,_Flag,VFlag,NFlag);
 sw=set of (s0,s1,s2,s3,s4,s5,s6,s7);
 TC64=Class
 // Data
  RAM  :Array[$0000..$FFFF] of byte;
  BASIC:Array[$A000..$BFFF] of byte; BasicOn:boolean;
  CHARS:Array[$D000..$DFFF] of byte; CharsOn:boolean; ChipsOn:boolean;  // Override CharsOn
  KERNL:Array[$E000..$FFFF] of byte; KernlOn:boolean;
  PC,A,X,Y,S:word; IR:byte;
  P:TFlags;
  VIC:TVIC;
  CIA1:TCIA1;
  CIA2:TCIA2;
  Cycles:integer;
 // Init...
  Constructor Create;
  Destructor  Destroy; override;
  Procedure   Reset;
 // Memory Access
  Function Peek(Addr:word):byte;
  Function Deek(Addr:word):word;
  Procedure Poke(Addr:word; value:byte);
  Procedure Push(b:byte);
  Function  Pop:byte;
  Procedure PushW(w:word);
 // Flags
  Procedure SetFlags(Flags:TFlags; OnOff:boolean);
  Function SetNZ(b:byte):byte;
 // Addressing modes
  Function Imm:byte;
  Function ZPge:byte;
  Function ZPgX:byte;
  Function ZPgY:byte;
  Function Abso:word;
  Function AbsX:word;
  Function AbsY:word;
  Function IndX:word;
  Function IndY:word;
  Function PeekNZ(Addr:word):byte;
  Function PeekZPge:byte;
  Function PeekNZZPge:byte;
 // OpCodes
  Procedure JAM;
  Procedure ADC(Value:byte);
  Function  ASL(b:byte):byte;
  Procedure ASL_At(Addr:word);
  Procedure BIT_At(Addr:word);
  Procedure Branch(Test:boolean);
  Procedure CMP(Reg,Value:byte);
  Procedure DEC_At(Addr:word);
  Procedure INC_At(Addr:word);
  Function  LSR(b:byte):byte;
  Procedure LSR_At(Addr:word);
  Function  ROL(b:byte):byte;
  Procedure ROL_At(Addr:word);
  Function  ROR(b:byte):byte;
  Procedure ROR_At(Addr:word);
  Procedure SBC(Value:byte);
 // Emulation
  Procedure EmulateLine;
  Procedure CheckKeyboard;
  Function Emulate6510:integer;
  Procedure NewConfig;
 {$IFDEF DEBUG}
  Function FlagStr:string;
  Procedure DumpRegs;
 {$ENDIF}
 {$IFDEF LIST}
  Procedure List;
 {$ENDIF}
 end;

var
 CBM64:TC64;

implementation

Constructor TC64.Create;
 begin
  VIC:=TVIC.Create;
  CIA1:=TCIA1.Create;
  CIA2:=TCIA2.Create;
  Reset;
  {$ifdef debug}
   dump:=false;
  {$endif}
  {$ifdef list}
   assignfile(lst,'CBM64W.LST'); rewrite(lst);
   assignfile(dos,'C:\TEMP\CBM64.LST'); system.reset(dos);
  {$endif}
 end;

Destructor TC64.Destroy;
 begin
  VIC.Free;
  inherited Destroy;
 end;

Procedure TC64.Reset;
 begin
  LoadData(BASIC,'BASIC_ROM');
  LoadData(CHARS,'CHARS_ROM');
  LoadData(KERNL,'KERNL_ROM');
  FillChar(RAM,SizeOf(RAM),0);
  Keybuffer:='';
  NewConfig;
  PC:=$FCE2; A:=0; X:=0; Y:=0; S:=$FF; P:=[_Flag];
  VIC.Reset;
  CIA1.Reset;
  CIA2.Reset;
  Cycles:=0;
 end;

Procedure TC64.EmulateLine;
 begin
{
  While Cycles>=0 do Dec(Cycles,Emulate6510);
}
  Dec(Cycles,Emulate6510);
  if Cycles<0 then begin
   Inc(Cycles,VIC.EmulateLine);
   CIA1.EmulateLine(Cycles);
   CIA2.EmulateLine(Cycles);
   if CIA2.INT then begin // NMI
     CIA2.INT:=False;
     PushW(PC);
     Push(byte(P));
     P:=P+[IFlag];
     PC:=Deek($FFFA);
     Dec(Cycles,7);
   end else if CIA1.INT then begin // IRQ
    {$ifdef lst}
      if (IFlag in P) then writeln(lst,'IRQ skipped') else writeln(lst,'IRQ');
    {$endif}
    if not (IFlag in P) then begin
     CIA1.INT:=False;
     PushW(PC);
     Push(byte(P));
     P:=P+[IFlag];
     PC:=Deek($FFFE);
     Dec(Cycles,7);
   end;
  end;
 end;
end; 


Procedure TC64.CheckKeyboard;
 var
  key:byte;
  c:word;
  col,row:word;
 begin
  FillChar(KeyMatrix,SizeOf(KeyMatrix),$FF);
  FillChar(RevMatrix,SizeOf(RevMatrix),$FF);
  if Keybuffer='' then exit;
  key:=ord(Keybuffer[length(Keybuffer)]);
  SetLength(Keybuffer,Length(Keybuffer)-1);
  if (key>0) and (key<128) then begin
 //  MessageBox(hWnd,pchar('Invalide key '+inttostr(key)),Title,0);  
   if GetKeyState(VK_SHIFT)<0 then c:=SKEY[key] else
   if GetKeyState(VK_MENU) <0 then c:=AKEY[key] else c:=KEYS[key];
   case c of
    $0E : halt;
    $0F : exit;
    $FF : MessageBox(hWnd,pchar('Invalide key '+inttostr(key)),Title,0);
    else  begin
     col:=(c shr 4) and 7;
     row:=c and 7;
     if c>255 then begin
      KeyMatrix[Col]:=255-(1 shl Row);
      RevMatrix[Row]:=255-(1 shl Col);
     { shift }
      KeyMatrix[1]:=KeyMatrix[1] and (255-(1 shl 7));
      RevMatrix[7]:=KeyMatrix[7] and (255-(1 shl 1));
     end else begin
      KeyMatrix[Col]:=255-(1 shl Row);
      RevMatrix[Row]:=255-(1 shl Col);
     end;
    end;
   end;
  end;
 end;


Function TC64.Peek(Addr:word):byte;
 begin
  Result:=0;
  Case Addr of
   $0000..$9FFF: Result:=RAM[Addr];
   $A000..$BFFF: if BasicOn then Result:=BASIC[Addr] else result:=RAM[Addr];
   $C000..$CFFF: Result:=RAM[Addr];
   $D000..$DFFF: if ChipsOn then
                  case (Addr and $FF00) of
                    $D000..$D300 : Result:=VIC.Peek(Addr);
                //    $D400..$D700 : SID.Poke(Addr and $1F,Value);
                    $D800..$DB00 : Result:=VIC.Color[Addr];
                    $DC00        : Result:=CIA1.Peek(Addr);
                    $DD00        : Result:=CIA2.Peek(Addr);
                 //   $DE00..$DF00 : if (Addr and $FFF0)=$DF00 then PokeREU(Addr and $F,Value);
                  end
                 else
                 if CharsOn then result:=CHARS[Addr] else result:=RAM[Addr];
   $E000..$FFFF: if KernlOn then Result:=KERNL[Addr] else Result:=RAM[Addr];
  end;
 end;

Function TC64.Deek(Addr:word):word;
 begin
  Result:=(Peek(Addr) or Peek(Addr+1) shl 8)
 end;

Procedure TC64.Poke(Addr:word; Value:byte);
 begin
  case Addr of
   $0000..$CFFF : begin RAM[Addr]:=Value; if Addr<2 then NewConfig end;
   $D000..$DFFF : if ChipsOn then begin
                   case (Addr and $FF00) of
                    $D000..$D300 : VIC.Poke(Addr,Value);
                //    $D400..$D700 : SID.Poke(Addr and $1F,Value);
                    $D800..$DB00 : VIC.Color[Addr]:=Value and $F;
                    $DC00        : CIA1.Poke(Addr,Value);
                    $DD00        : CIA2.Poke(Addr,Value);
                 //   $DE00..$DF00 : if (Addr and $FFF0)=$DF00 then PokeREU(Addr and $F,Value);
                   end;
                  end else
                   RAM[Addr]:=Value;
   $E000..$FFFF : begin RAM[Addr]:=Value; {if Addr=$FF00 then REUFF00} end;
  end;
 end;

Procedure TC64.Push(b:byte);
 begin
  RAM[S or $100]:=b; S:=Pred(S) and $FF;
 end;

Function TC64.Pop:byte;
 begin
  S:=Succ(S) and $FF; Result:=RAM[S or $100];
 end;

Procedure TC64.PushW(w:word);
 begin
  RAM[S or $100]:=hi(w); if s=0 then s:=$FF else dec(s);
  RAM[S or $100]:=lo(w); if s=0 then s:=$FF else dec(s);
 end;

Procedure TC64.SetFlags(Flags:TFlags; OnOff:boolean);
 begin
  if OnOff then P:=P+Flags else P:=P-Flags;
 end;

Function TC64.SetNZ(b:byte):byte;
 begin
  SetFlags([NFlag],b>127);
  SetFlags([ZFlag],b=0);
  Result:=b;
 end;

Function TC64.Imm:byte;      begin Result:=Peek(PC); inc(PC); end;
Function TC64.Zpge:byte;     begin Result:=Peek(PC); inc(PC); end;
Function TC64.ZPgX:byte;     begin Result:=(ZPge+X) and 255; end;
Function TC64.ZPgY:byte;     begin Result:=ZPge+Y; end;
Function TC64.Abso:word;     begin Result:=Imm; Result:=Result or (Imm shl 8); end;
Function TC64.AbsX:word;     begin Result:=Imm; Result:=Result or (Imm shl 8)+X; end;
Function TC64.AbsY:word;     begin Result:=Imm; Result:=Result or (Imm shl 8)+Y; end;
Function TC64.IndX:word;     begin Result:=Deek(Imm+X); end;
Function TC64.IndY:word;     begin Result:=Deek(Imm)+Y; end;

Function TC64.PeekNZ(Addr:word):byte; begin Result:=SetNZ(Peek(Addr)) end;
Function TC64.PeekZPge:byte; begin Result:=RAM[Imm]; end;
Function TC64.PeekNZZPge:byte; begin Result:=SetNZ(RAM[Imm]); end;

Procedure TC64.JAM;
 begin
  case MessageBox(hWnd,JamError,Title,MB_ICONERROR or MB_YESNOCANCEL) of
   idYes    : Reset;
   idCancel : Halt;
  end;
 end;

Procedure TC64.ADC(Value:byte);
 var
  som:word;
 begin
  som:=A+Value; if CFlag in P then inc(som);
  SetFlags([CFlag],(Som>255));
  SetFlags([VFlag],boolean((not (a xor value) and $80) and ((a xor som) and  $80)));
  A:=SetNZ(som and $ff);
 end;

Function  TC64.ASL(b:byte):byte;  begin SetFlags([CFlag],b>127); Result:=SetNZ((b shl 1) and $ff); end;
Procedure TC64.ASL_At(Addr:word); begin Poke(Addr,ASL(Peek(Addr))); end;

Procedure TC64.BIT_At(Addr:word);
 Var
  M:Byte;
 begin
  M:=Peek(Addr);
  SetFlags([ZFlag],(A and M)=0);
  SetFlags([VFlag],s6 in sw(M));
  SetFlags([NFlag],s7 in sw(M));
 end;

Procedure TC64.Branch(Test:boolean);
 var rela:shortint;
 begin
  rela:=ShortInt(Imm);
  if Test then PC:=(PC+Rela) and $FFFF;
 end;

Procedure TC64.CMP(Reg,Value:byte);
 begin
//  SetFlags([NFlag],Reg<Value);
  SetFlags([NFlag],(Reg-Value) and 128=128);
  SetFlags([ZFlag],Reg=Value);
  SetFlags([CFlag],Reg>=Value);
 end;

Procedure TC64.DEC_at(Addr:word);
 begin
  Poke(Addr,SetNZ(Pred(Peek(Addr)) and $FF));
 end;

Procedure TC64.INC_at(Addr:word);
 begin
  Poke(Addr,SetNZ(Succ(Peek(Addr)) and $FF));
 end;

Function TC64.LSR(b:byte):byte;
 begin
  SetFlags([CFlag],odd(b));
  LSR:=SetNZ(B SHR 1);
 end;

Procedure TC64.LSR_at(Addr:word); begin Poke(Addr,LSR(Peek(Addr))); end;

Function TC64.ROL(b:byte):byte;
 Var
  carry:boolean;
 begin
  carry:=CFlag in P;
  SetFlags([CFlag],b>127);
  b:=(b SHL 1) and $FF; if carry then b:=b or 1;
  Result:=SetNZ(b);
 end;

Procedure TC64.ROL_at(Addr:word); begin Poke(Addr,ROL(Peek(Addr))); end;

Function TC64.ROR(b:byte):byte;
 var
  carry:boolean;
 begin
  carry:=CFlag in P;
  SetFlags([CFlag],odd(b));
  b:=b SHR 1;
  if carry then b:=b or 128;
  ROR:=SetNZ(b);
 end;

Procedure TC64.ROR_at(Addr:word); begin Poke(Addr,ROR(Peek(Addr))); end;

Procedure TC64.SBC(Value:byte);
 var
  sub:Integer;
 begin
  sub:=A-Value; if not (CFlag in P) then dec(sub);
  SetFlags([CFlag],(Sub>=0));
//  SetFlags([VFlag],(A and 128)<>(Sub and 128));
//  v_flag = ((a ^ tmp) & 0x80) && ((a ^ byte) & 0x80);
  SetFlags([VFlag],boolean(((a xor value) and $80) and ((a xor sub) and  $80)));
  A:=SetNZ(sub and $FF);
 end;

Function TC64.Emulate6510:integer;
 begin
  Result:=0;
  try
   IR:=Peek(PC);

   IF A>255 then MessageBox(hWnd,'A overflow',Title,0);
   IF X>255 then MessageBox(hWnd,'X overflow',Title,0);
   IF Y>255 then MessageBox(hWnd,'Y overflow',Title,0);
   IF S>255 then MessageBox(hWnd,'S overflow',Title,0);
  {$IFDEF DEBUG}
//    if PC=$FD25 then dump:=true;  // Badline condition bug - Inside KERNEL setup
    if PC=$E394 then dump:=true; // IRQ Timing bug - KERNEL setup done, now BASIC setup
//    dump:=true; // CIA Timer bug
    if dump then DumpRegs;
  {$ENDIF}
  {$IFDEF LIST}
    List;
  {$ENDIF}
   Inc(PC); // Get OpCode
  case IR of
   $00 { BRK      } : begin
                       PushW(PC+1);
                       Push(Byte(P+[BFlag]));
                       PC:=Deek($FFFE);
                       Result:=7
                      end;
   $01 { ORA IndX } : begin A:=SetNZ(A or Peek(IndX)); Result:=6 end;
   $05 { ORA ZPge } : begin A:=SetNZ(A or PeekZPge); Result:=3 end;
   $06 { ASL ZPge } : begin ASL_at(Imm); Result:=5 end;
   $08 { PHP      } : begin push(byte(P)); Result:=3 end;
   $09 { ORA Imm  } : begin A:=SetNZ(A or Imm); Result:=2 end;
   $0A { ASL      } : begin A:=ASL(A); Result:=2 end;
   $0D { ORA Abso } : begin A:=SetNZ(A or Peek(Abso)); Result:=4 end;
   $0E { ASL Abso } : begin ASL_at(Abso); Result:=6 end;
   $10 { BPL Rela } : begin Branch(Not (NFlag in P)); Result:=2 end;
   $11 { ORA IndY } : begin A:=SetNZ(A or Peek(IndY)); Result:=5 end;
   $15 { ORA ZPgX } : begin A:=SetNZ(A or Peek(ZPgX)); Result:=4 end;
   $16 { ASL ZPgX } : begin ASL_at(ZPgX); Result:=6 end;
   $18 { CLC      } : begin P:=P-[CFlag]; Result:=2 end;
   $19 { ORA AbsY } : begin A:=SetNZ(A or Peek(AbsY)); Result:=4 end;
   $1D { ORA AbsX } : begin A:=SetNZ(A or Peek(AbsX)); Result:=4 end;
   $1E { ASL AbsX } : begin ASL_at(AbsX); Result:=7 end;
   $20 { JSR Abso } : begin PushW(PC+1); PC:=Abso; result:=6 end;
   $21 { AND IndX } : begin A:=SetNZ(A and Peek(IndX)); result:=6 end;
   $24 { BIT ZPge } : begin BIT_at(ZPge); Result:=3 end;
   $25 { AND ZPge } : begin A:=SetNZ(A and PeekZPge); Result:=3 end;
   $26 { ROL ZPge } : begin ROL_at(Imm); Result:=5 end;
   $28 { PLP      } : begin byte(P):=pop; Result:=4 end;
   $29 { AND Imm  } : begin A:=SetNZ(A and Imm); result:=2 end;
   $2A { ROL      } : begin A:=ROL(A); result:=2 end;
   $2C { BIT Abso } : begin BIT_at(Abso); result:=4 end;
   $2D { AND Abso } : begin A:=SetNZ(A and Peek(Abso)); result:=4 end;
   $2E { ROL Abso } : begin ROL_at(Abso);  result:=6 end;
   $30 { BMI Rela } : begin Branch(NFlag in P);  result:=2 end;
   $31 { AND IndY } : begin A:=SetNZ(A and Peek(IndY)); result:=4 end;
   $35 { AND ZPgX } : begin A:=SetNZ(A and Peek(ZPgX)); result:=4 end;
   $36 { ROL ZPgX } : begin ROL_at(ZPgX); result:=6 end;
   $38 { SEC      } : begin P:=P+[CFlag]; result:=2 end;
   $39 { AND AbsY } : begin A:=SetNZ(A and Peek(AbsY)); result:=4 end;
   $3D { AND AbsX } : begin A:=SetNZ(A and Peek(AbsX)); result:=4 end;
   $3E { ROL AbsX } : begin ROL_at(AbsX); result:=7 end;
   $40 { RTI      } : begin byte(P):=pop; PC:=pop+pop shl 8; result:=6 end;
   $41 { EOR IndX } : begin A:=SetNZ(A xor Peek(IndX)); result:=6 end;
   $45 { EOR ZPge } : begin A:=SetNZ(A xor PeekZPge); result:=3 end;
   $46 { LSR ZPge } : begin LSR_at(Imm); result:=5 end;
   $48 { PHA      } : begin Push(A); result:=3 end;
   $49 { EOR Imm  } : begin A:=SetNZ(A xor Imm); result:=2 end;
   $4A { LSR      } : begin A:=LSR(A); result:=2 end;
   $4C { JMP Abso } : begin PC:=Abso; result:=3 end;
   $4D { EOR Abso } : begin A:=SetNZ(A xor Peek(Abso)); result:=4 end;
   $4E { LSR Abso } : begin LSR_at(Abso); result:=6 end;
   $50 { BVC      } : begin Branch(Not(VFlag in P)); result:=2 end;
   $51 { EOR IndY } : begin A:=SetNZ(A xor Peek(IndY)); result:=5 end;
   $55 { EOR ZPgX } : begin A:=SetNZ(A xor Peek(ZPgX)); result:=4 end;
   $56 { LSR ZPgX } : begin LSR_at(ZPgX); result:=6 end;
   $58 { CLI      } : begin Exclude(P,IFlag); result:=2 end;
   $59 { EOR AbsY } : begin A:=SetNZ(A xor Peek(AbsY)); result:=4 end;
   $5D { EOR AbsX } : begin A:=SetNZ(A xor Peek(AbsX)); result:=4 end;
   $5E { LSR AbsX } : begin LSR_at(AbsX); result:=7 end;
   $60 { RTS      } : begin PC:=(pop+pop shl 8)+1; result:=6 end;
   $61 { ADC IndX } : begin ADC(Peek(IndX)); result:=6 end;
   $65 { ADC ZPge } : begin ADC(PeekZPge); result:=3 end;
   $66 { ROR ZPge } : begin ROR_at(ZPge); result:=5 end;
   $68 { PLA      } : begin A:=SetNZ(POP); result:=4 end;
   $69 { ADC Imm  } : begin ADC(Imm); result:=2 end;
   $6A { ROR      } : begin A:=ROR(A); result:=2 end;
   $6C { JMP Ind  } : begin PC:=Deek(Deek(PC)); result:=5 end;
   $6D { ADC Abso } : begin ADC(Peek(Abso)); result:=4 end;
   $6E { ROR Abso } : begin ROR_at(Abso); result:=6 end;
   $70 { BVS      } : begin Branch(VFlag in P); result:=2 end;
   $71 { ADC IndY } : begin ADC(Peek(IndY)); result:=5 end;
   $75 { ADC ZPgX } : begin ADC(Peek(ZPgX)); result:=4 end;
   $76 { ROR ZPgX } : begin ROR_at(ZPgX); result:=6 end;
   $78 { SEI      } : begin P:=P+[IFlag]; result:=2 end;
   $79 { ADC AbsY } : begin ADC(Peek(AbsY)); result:=4 end;
   $7D { ADC AbsX } : begin ADC(Peek(AbsX)); result:=4 end;
   $7E { ROR AbsX } : begin ROR_at(AbsX); result:=7 end;
   $81 { STA IndX } : begin Poke(IndX,A); result:=6 end;
   $84 { STY ZPge } : begin Poke(ZPge,Y); result:=3 end;
   $85 { STA ZPge } : begin Poke(ZPge,A); result:=3 end;
   $86 { STX ZPge } : begin Poke(ZPge,X); result:=3 end;
   $88 { DEY      } : begin if Y=0 then Y:=SetNZ($FF) else Y:=SetNZ(Y-1); result:=2 end;
   $8A { TXA      } : begin A:=SetNZ(X); result:=2 end;
   $8C { STY Abso } : begin Poke(Abso,Y); result:=4 end;
   $8D { STA Abso } : begin Poke(Abso,A); result:=4 end;
   $8E { STX Abso } : begin Poke(Abso,X); result:=4 end;
   $90 { BCC Rela } : begin Branch(Not (CFlag in P)); result:=2 end;
   $91 { STA IndY } : begin Poke(IndY,A); result:=5 end;
   $94 { STY ZPgX } : begin Poke(ZPgX,Y); result:=4 end;
   $95 { STA ZPgX } : begin Poke(ZPgX,A); result:=4 end;
   $96 { STX ZPgX } : begin Poke(ZPgX,X); result:=4 end;
   $98 { TYA      } : begin A:=SetNZ(Y); result:=2 end;
   $99 { STA AbsY } : begin Poke(AbsY,A); result:=4 end;
   $9A { TXS      } : begin S:=X; result:=2 end;
   $9D { STA AbsX } : begin Poke(AbsX,A); result:=4 end;
   $A0 { LDY Imm  } : begin Y:=SetNZ(Imm); result:=2 end;
   $A1 { LDA IndX } : begin A:=PeekNZ(IndX); result:=6 end;
   $A2 { LDX Imm  } : begin X:=SetNZ(Imm); result:=2 end;
   $A4 { LDY ZPge } : begin Y:=PeekNZZPge; result:=3 end;
   $A5 { LDA ZPge } : begin A:=PeekNZZPge; result:=3 end;
   $A6 { LDX ZPge } : begin X:=PeekNZZPge; result:=3 end;
   $A8 { TAY      } : begin Y:=SetNZ(A); result:=2 end;
   $A9 { LDA Imm  } : begin A:=SetNZ(Imm); result:=2 end;
   $AA { TAX      } : begin X:=SetNZ(A); result:=2 end;
   $AC { LDY Abso } : begin Y:=PeekNZ(Abso); result:=4 end;
   $AD { LDA Abso } : begin A:=PeekNZ(Abso); result:=4 end;
   $AE { LDX Abso } : begin X:=PeekNZ(Abso); result:=4 end;
   $B0 { BCS Rela } : begin Branch(CFlag in P); result:=2 end;
   $B1 { LDA IndY } : begin A:=PeekNZ(IndY); result:=5 end;
   $B4 { LDY ZPgX } : begin Y:=PeekNZ(ZPgX); result:=4 end;
   $B5 { LDA ZPgX } : begin A:=PeekNZ(ZPgX); result:=4 end;
   $B6 { LDX ZPgY } : begin X:=PeekNZ(ZPgY); result:=4 end;
   $B8 { CLV      } : begin Exclude(P,VFlag); result:=2 end;
   $B9 { LDA AbsY } : begin A:=PeekNZ(AbsY); result:=4 end;
   $BA { TSX      } : begin X:=SetNZ(S); result:=2 end;
   $BC { LDY AbsX } : begin Y:=PeekNZ(AbsX); result:=4 end;
   $BD { LDA AbsX } : begin A:=PeekNZ(AbsX); result:=4 end;
   $BE { LDX AbsY } : begin X:=PeekNZ(AbsY); result:=4 end;
   $C0 { CPY Imm  } : begin CMP(Y,Imm); result:=2 end;
   $C1 { CMP IndX } : begin CMP(A,Peek(IndX)); result:=6 end;
   $C4 { CPY ZPge } : begin CMP(Y,PeekZPge); result:=3 end;
   $C5 { CMP ZPge } : begin CMP(A,PeekZPge); result:=3 end;
   $C6 { DEC ZPge } : begin DEC_at(ZPge); result:=5 end;
   $C8 { INY      } : begin if Y=$FF then Y:=SetNZ(0) else Y:=SetNZ(Y+1); result:=2 end;
   $C9 { CMP Imm  } : begin CMP(A,Imm); result:=2 end;
   $CA { DEX      } : begin if x=0 then X:=SetNZ($FF) else X:=SetNZ(X-1); result:=2 end;
   $CC { CPY Abso } : begin CMP(Y,Peek(Abso)); result:=4 end;
   $CD { CMP Abso } : begin CMP(A,Peek(Abso)); result:=4 end;
   $CE { DEC Abso } : begin DEC_at(Abso); result:=6 end;
   $D0 { BNE Rela } : begin Branch(not (ZFlag in P)); result:=2 end;
   $D1 { CMP IndY } : begin CMP(A,Peek(IndY)); result:=5 end;
   $D5 { CMP ZPgX } : begin CMP(A,Peek(ZPgX)); result:=4 end;
   $D6 { DEC ZPgX } : begin DEC_at(ZPgX); result:=6 end;
   $D8 { CLD      } : begin Exclude(P,CFlag); result:=2 end;
   $D9 { CMP AbsY } : begin CMP(A,Peek(AbsY)); result:=4 end;
   $DD { CMP AbsX } : begin CMP(A,Peek(AbsX)); result:=4 end;
   $DE { DEC AbsX } : begin DEC_at(AbsX); result:=7 end;
   $E0 { CPX Imm  } : begin CMP(X,Imm); result:=2 end;
   $E1 { SBC IndX } : begin SBC(Peek(IndX)); result:=6 end;
   $E4 { CPX ZPge } : begin CMP(X,PeekZPge); result:=3 end;
   $E5 { SBC ZPge } : begin SBC(PeekZPge); result:=3 end;
   $E6 { INC ZPge } : begin INC_at(ZPge); result:=5 end;
   $E8 { INX      } : begin X:=SetNZ((X+1) and $FF); result:=2 end;
   $E9 { SBC Imm  } : begin SBC(Imm); result:=2 end;
   $EA { NOP      } : begin { nothing ! }; result:=2 end;
   $EC { CPX Abso } : begin CMP(X,Peek(Abso)); result:=4 end;
   $ED { SBC Abso } : begin SBC(Peek(Abso)); result:=4 end;
   $EE { INC Abso } : begin INC_at(Abso); result:=6 end;
   $F0 { BEQ Rela } : begin Branch(ZFlag in P); result:=2 end;
   $F1 { SBC IndY } : begin SBC(Peek(IndY)); result:=5 end;
(*
   $F2 { Extension OpCode } : begin
    if PC<$E000 then Halt(hJAM);
    Case Imm of
     $00 : begin
            Bank0^[$90]:=Bank0^[$90] or IEC.Output(Bank0^[$95],Bank0^[$A3] and $80);
            Exclude(P,CFg);
            PC:=$EDAC;
           end;
     $01 : begin
            Bank0^[$90]:=Bank0^[$90] or IEC.OutputATN(A{Bank0^[$95]});
            Exclude(P,CFg);
            PC:=$EDAC;
           end;

     $02 : begin
            Bank0^[$90]:=Bank0^[$90] or IEC.OutputSec(Bank0^[$95]);
            Exclude(P,CFg);
            PC:=$EDAC;
           end;

     $03 : begin
            Bank0^[$90]:=IEC.Input(A);
            SetNZ(A);
            Exclude(P,CFg);
            PC:=$EDAC;
           end;
     else  Halt(hIEC);
    end;
   end;
*)
   $F5 { SBC ZPgX } : begin SBC(Peek(ZPgX)); result:=4 end;
   $F6 { INC ZPgX } : begin INC_at(ZPgX); result:=6 end;
//   $F8 { SED      } : begin Halt(hSED);
   $F9 { SBC AbsY } : begin SBC(Peek(AbsY)); result:=4 end;
   $FD { SBC AbsX } : begin SBC(Peek(AbsX)); result:=4 end;
   $FE { INC AbsX } : begin INC_at(AbsX); result:=7 end;
   else JAM;
  end;
  except
    messagebox(hwnd,pchar('Error on OP='+IntToHex(IR,2)),title,0);
  end;
 end;

Procedure TC64.NewConfig;
 Var
  Port:byte;
 begin
  Port:=(not RAM[0]) or RAM[1];
  KernlOn:=(Port and 2)=2;
  BasicOn:=(Port and 3)=3;
  CharsOn:=(Port and 3)>0;
  ChipsOn:=(Port and 7)>4; // override CharsOn
 end;
{------------------------------------------------------------------------------}
{$IFDEF DEBUG}
Type
 TParm=(
  p0,
  pImpl, { no parameters }
  pAbso, { word }
  pZPge, { byte }
  pImm,  { #byte }
  pAbsX, { word+x }
  pAbsY, { word+y }
  pIndX, { [byte+x] = array[x]^ }
  pIndY, { [byte]+y = array^[y] }
  pZPgX, { byte+x }
  pZPgY, { byte+y }
  pRela, { shortint(byte) }
  pInd   { [word] }
 );

 TOpCode=record
  Name:array[1..3] of char;
  Parm:TParm;
 end;

Const
 OpCode:Array[0..255] of TOpCode = (
 {00-0F}
  (Name:'BRK'; Parm:pImpl),  (Name:'ORA'; Parm:pIndX),
  (Name:'JAM'; Parm:pImpl),  (Name:'SLO'; Parm:pIndX),
  (Name:'NOP'; Parm:pZpge),  (Name:'ORA'; Parm:pZPge),
  (Name:'ASL'; Parm:pZPge),  (Name:'SLO'; Parm:pZPge),
  (Name:'PHP'; Parm:pImpl),  (Name:'ORA'; Parm:pImm),
  (Name:'ASL'; Parm:pImpl),  (Name:'ANC'; Parm:pImm),
  (Name:'NOP'; Parm:pAbso),  (Name:'ORA'; Parm:pAbso),
  (Name:'ASL'; Parm:pAbso),  (Name:'   '; Parm:p0   ),
 {10-1F}
  (Name:'BPL'; Parm:pRela),  (Name:'ORA'; Parm:pIndY),
  (Name:'   '; Parm:p0   ),  (Name:'SLO'; Parm:pIndY),
  (Name:'NOP'; Parm:pZPgX),  (Name:'ORA'; Parm:pZPgX),
  (Name:'ASL'; Parm:pZPgX),  (Name:'   '; Parm:p0   ),
  (Name:'CLC'; Parm:pImpl),  (Name:'ORA'; Parm:pAbsY),
  (Name:'   '; Parm:p0   ),  (Name:'   '; Parm:p0   ),
  (Name:'   '; Parm:p0   ),  (Name:'ORA'; Parm:pAbsX),
  (Name:'ASL'; Parm:pAbsX),  (Name:'   '; Parm:p0   ),
 {20-2F}
  (Name:'JSR'; Parm:pAbso),  (Name:'AND'; Parm:pIndX),
  (Name:'   '; Parm:p0   ),  (Name:'RLA'; Parm:pIndX),
  (Name:'BIT'; Parm:pZPge),  (Name:'AND'; Parm:pZPge),
  (Name:'ROL'; Parm:pZPge),  (Name:'RLA'; Parm:pZPge),
  (Name:'PLP'; Parm:pImpl),  (Name:'AND'; Parm:pImm),
  (Name:'ROL'; Parm:pImpl),  (Name:'   '; Parm:p0   ),
  (Name:'BIT'; Parm:pAbso),  (Name:'AND'; Parm:pAbso),
  (Name:'ROL'; Parm:pAbso),  (Name:'RLA'; Parm:pAbso),
 {30-3F}
  (Name:'BMI'; Parm:pRela),  (Name:'AND'; Parm:pIndY),
  (Name:'   '; Parm:p0   ),  (Name:'RLA'; Parm:pIndY),
  (Name:'   '; Parm:p0   ),  (Name:'AND'; Parm:pZPgX),
  (Name:'ROL'; Parm:pZPgX),  (Name:'RLA'; Parm:pZPgX),
  (Name:'SEC'; Parm:pImpl),  (Name:'AND'; Parm:pAbsY),
  (Name:'   '; Parm:p0   ),  (Name:'RLA'; Parm:pAbsY),
  (Name:'   '; Parm:p0   ),  (Name:'AND'; Parm:pAbsX),
  (Name:'ROL'; Parm:pAbsX),  (Name:'RLA'; Parm:pAbsX),
 {40-4F}
  (Name:'RTI'; Parm:pImpl),  (Name:'EOR'; Parm:pIndX),
  (Name:'   '; Parm:p0   ),  (Name:'SRE'; Parm:pIndX),
  (Name:'   '; Parm:p0   ),  (Name:'EOR'; Parm:pZPge),
  (Name:'LSR'; Parm:pZPge),  (Name:'   '; Parm:p0   ),
  (Name:'PHA'; Parm:pImpl),  (Name:'EOR'; Parm:pImm ),
  (Name:'LSR'; Parm:pImpl),  (Name:'ASR'; Parm:pImm ),
  (Name:'JMP'; Parm:pAbso),  (Name:'EOR'; Parm:pAbso),
  (Name:'LSR'; Parm:pAbso),  (Name:'SRE'; Parm:pAbso),
 {50-5F}
  (Name:'BVC'; Parm:pRela),  (Name:'EOR'; Parm:pIndY),
  (Name:'   '; Parm:p0   ),  (Name:'SRE'; Parm:pIndY),
  (Name:'   '; Parm:p0   ),  (Name:'EOR'; Parm:pZPgX),
  (Name:'LSR'; Parm:pZPgX),  (Name:'SRE'; Parm:pZPgX),
  (Name:'CLI'; Parm:pImpl),  (Name:'EOR'; Parm:pAbsY),
  (Name:'   '; Parm:p0   ),  (Name:'SRE'; Parm:pAbsY),
  (Name:'   '; Parm:p0   ),  (Name:'EOR'; Parm:pAbsX),
  (Name:'LSR'; Parm:pAbsX),  (Name:'SRE'; Parm:pAbsX),
 {60-6F}
  (Name:'RTS'; Parm:pImpl),  (Name:'ADC'; Parm:pIndX),
  (Name:'   '; Parm:p0   ),  (Name:'RRA'; Parm:pIndX),
  (Name:'   '; Parm:p0   ),  (Name:'ADC'; Parm:pZPge),
  (Name:'ROR'; Parm:pZPge),  (Name:'RRA'; Parm:pZPge),
  (Name:'PLA'; Parm:pImpl),  (Name:'ADC'; Parm:pImm),
  (Name:'ROR'; Parm:pImpl),  (Name:'   '; Parm:p0   ),
  (Name:'JMP'; Parm:pInd ),  (Name:'ADC'; Parm:pAbso),
  (Name:'ROR'; Parm:pAbso),  (Name:'RRA'; Parm:pAbso),
 {70-7F}
  (Name:'BVS'; Parm:pRela),  (Name:'ADC'; Parm:pIndY),
  (Name:'   '; Parm:p0   ),  (Name:'RRA'; Parm:pIndY),
  (Name:'   '; Parm:p0   ),  (Name:'ADC'; Parm:pZPgX),
  (Name:'ROR'; Parm:pZPgX),  (Name:'RRA'; Parm:pZPgX),
  (Name:'SEI'; Parm:pImpl),  (Name:'ADC'; Parm:pAbsY),
  (Name:'   '; Parm:p0   ),  (Name:'RRA'; Parm:pAbsY),
  (Name:'   '; Parm:p0   ),  (Name:'ADC'; Parm:pAbsX),
  (Name:'ROR'; Parm:pAbsX),  (Name:'RRA'; Parm:pAbsX),
 {80-8F}
  (Name:'NOP'; Parm:pImm ),  (Name:'STA'; Parm:pIndX),
  (Name:'   '; Parm:p0   ),  (Name:'SAX'; Parm:pIndX),
  (Name:'STY'; Parm:pZPge),  (Name:'STA'; Parm:pZPge),
  (Name:'STX'; Parm:pZPge),  (Name:'SAX'; Parm:pZPge),
  (Name:'DEY'; Parm:pImpl),  (Name:'   '; Parm:p0   ),
  (Name:'TXA'; Parm:pImpl),  (Name:'ANE'; Parm:pImm ),
  (Name:'STY'; Parm:pAbso),  (Name:'STA'; Parm:pAbso),
  (Name:'STX'; Parm:pAbso),  (Name:'SAX'; Parm:pAbso),
 {90-9F}
  (Name:'BCC'; Parm:pRela),  (Name:'STA'; Parm:pIndY),
  (Name:'   '; Parm:p0   ),  (Name:'SHA'; Parm:pIndY),
  (Name:'STY'; Parm:pZPgX),  (Name:'STA'; Parm:pZPgX),
  (Name:'STX'; Parm:pZPgY),  (Name:'SAX'; Parm:pZpgX),
  (Name:'TYA'; Parm:pImpl),  (Name:'STA'; Parm:pAbsY),
  (Name:'TXS'; Parm:pImpl),  (Name:'SHS'; Parm:pAbsY),
  (Name:'SHY'; Parm:pAbsX),  (Name:'STA'; Parm:pAbsX),
  (Name:'SHX'; Parm:pAbsY),  (Name:'SHA'; Parm:pAbsY),
 {A0-AF}
  (Name:'LDY'; Parm:pImm),   (Name:'LDA'; Parm:pIndX),
  (Name:'LDX'; Parm:pImm),   (Name:'LAX'; Parm:pIndX),
  (Name:'LDY'; Parm:pZPge),  (Name:'LDA'; Parm:pZPge),
  (Name:'LDX'; Parm:pZPge),  (Name:'LAX'; Parm:pZPge),
  (Name:'TAY'; Parm:pImpl),  (Name:'LDA'; Parm:pImm),
  (Name:'TAX'; Parm:pImpl),  (Name:'LXA'; Parm:pImm ),
  (Name:'LDY'; Parm:pAbso),  (Name:'LDA'; Parm:pAbso),
  (Name:'LDX'; Parm:pAbso),  (Name:'LAX'; Parm:pAbso),
 {B0-BF}
  (Name:'BCS'; Parm:pRela),  (Name:'LDA'; Parm:pIndY),
  (Name:'   '; Parm:p0   ),  (Name:'LAX'; Parm:pIndY),
  (Name:'LDY'; Parm:pZPgX),  (Name:'LDA'; Parm:pZpgX),
  (Name:'LDX'; Parm:pZPgY),  (Name:'LAX'; Parm:pZpgY),
  (Name:'CLV'; Parm:pImpl),  (Name:'LDA'; Parm:pAbsY),
  (Name:'TSX'; Parm:pImpl),  (Name:'LAS'; Parm:pAbsY),
  (Name:'LDY'; Parm:pAbsX),  (Name:'LDA'; Parm:pAbsX),
  (Name:'LDX'; Parm:pAbsY),  (Name:'LAX'; Parm:pAbsY),
 {C0-CF}
  (Name:'CPY'; Parm:pImm ),  (Name:'CMP'; Parm:pIndX),
  (Name:'   '; Parm:p0   ),  (Name:'DCP'; Parm:pIndX),
  (Name:'CPY'; Parm:pZPge),  (Name:'CMP'; Parm:pZPge),
  (Name:'DEC'; Parm:pZPge),  (Name:'DCP'; Parm:pZPge),
  (Name:'INY'; Parm:pImpl),  (Name:'CMP'; Parm:pImm),
  (Name:'DEX'; Parm:pImpl),  (Name:'SBX'; Parm:pImm ),
  (Name:'CPY'; Parm:pAbso),  (Name:'CMP'; Parm:pAbso),
  (Name:'DEC'; Parm:pAbso),  (Name:'DCP'; Parm:pAbso),
 {D0-DF}
  (Name:'BNE'; Parm:pRela),  (Name:'CMP'; Parm:pIndY),
  (Name:'   '; Parm:p0   ),  (Name:'DCP'; Parm:pIndY),
  (Name:'   '; Parm:p0   ),  (Name:'   '; Parm:p0   ),
  (Name:'DEC'; Parm:pZPgX),  (Name:'DCP'; Parm:pZPgX),
  (Name:'CLD'; Parm:pImpl),  (Name:'CMP'; Parm:pAbsY),
  (Name:'   '; Parm:p0   ),  (Name:'DCP'; Parm:pAbsY),
  (Name:'   '; Parm:p0   ),  (Name:'CMP'; Parm:pAbsX),
  (Name:'DEC'; Parm:pAbsX),  (Name:'DCP'; Parm:pAbsX),
 {E0-EF}
  (Name:'CPX'; Parm:pImm ),  (Name:'SBC'; Parm:pIndX),
  (Name:'   '; Parm:p0   ),  (Name:'ISB'; Parm:pIndX),
  (Name:'CPX'; Parm:pZPge),  (Name:'SBC'; Parm:pZPge),
  (Name:'INC'; Parm:pZPge),  (Name:'ISB'; Parm:pZPge),
  (Name:'INX'; Parm:pImpl),  (Name:'SBC'; Parm:pImm),
  (Name:'NOP'; Parm:pImpl),  (Name:'   '; Parm:p0   ),
  (Name:'CPX'; Parm:pAbso),  (Name:'SBC'; Parm:pAbso),
  (Name:'INC'; Parm:pAbso),  (Name:'ISB'; Parm:pAbso),
 {F0-FF}
  (Name:'BEQ'; Parm:pRela),  (Name:'SBC'; Parm:pIndY),
  (Name:'REU'; Parm:pImm ),  (Name:'ISB'; Parm:pIndY),
  (Name:'   '; Parm:p0   ),  (Name:'SBC'; Parm:pZPgX),
  (Name:'INC'; Parm:pZPgX),  (Name:'ISB'; Parm:pZPgX),
  (Name:'SED'; Parm:pImpl),  (Name:'SBC'; Parm:pAbsY),
  (Name:'   '; Parm:p0   ),  (Name:'ISB'; Parm:pAbsY),
  (Name:'   '; Parm:p0   ),  (Name:'SBC'; Parm:pAbsX),
  (Name:'INC'; Parm:pAbsX),  (Name:'ISB'; Parm:pAbsX)
  );

  Hexa:array[0..$F] of char='0123456789ABCDEF';

function HexaB(b:byte):string;
 begin
  Result:=Hexa[b shr 4]+Hexa[b and 15];
 end;

function HexaW(w:word):string;
 begin
  Result:=HexaB(w shr 8)+HexaB(w and 255);
 end;

Function TC64.FlagStr:string;
 begin
  Result:='';
  if NFlag in P then Result:=Result+('N') else Result:=Result+('n');
  if VFlag in P then Result:=Result+('V') else Result:=Result+('v');
  if _Flag in P then Result:=Result+('1') else Result:=Result+('0');
  if BFlag in P then Result:=Result+('B') else Result:=Result+('b');
  if DFlag in P then Result:=Result+('D') else Result:=Result+('d');
  if IFlag in P then Result:=Result+('I') else Result:=Result+('i');
  if ZFlag in P then Result:=Result+('Z') else Result:=Result+('z');
  if CFlag in P then Result:=Result+('C') else Result:=Result+('c');
 end;

Procedure TC64.DumpRegs;
 var
  pip2:word;
  pip1:byte absolute pip2;
  str:shortstring;

  function Peek(Addr:word):byte; // WARNING ! we don't want to reset CHIPS !
   begin
    case addr of
     $0000..$9FFF:Result:=RAM[Addr];
     $A000..$BFFF:Result:=BASIC[Addr];
     $C000..$CFFF:Result:=RAM[Addr];
     $D000..$D3FF:Result:=VIC.Regs.Raw[Addr and $3F];
     $D400..$D7FF:Result:=RAM[Addr]; // SID
     $D800..$DBFF:Result:=VIC.Color[Addr];
     $DC00..$DCFF:Result:=CIA1.Regs.Raw[Addr and $F];
     $DD00..$DDFF:Result:=CIA2.Regs.Raw[Addr and $F];
     $DE00..$DFFF:Result:=RAM[Addr]; // REU
     $E000..$FFFF:Result:=KERNL[Addr];
     else result:=$FF;
    end;
   end;

  Function Value(Addr:word):string;
   begin
    Result:=' ('+HexaW(Addr)+')= $'+HexaB(Peek(Addr));
   end;

 begin
  pip2:=Deek(PC+1);
  case PC of
   $A408 : WriteLn(#16' Check Memory');
   $AB1E : WriteLn(#16' Print string at AY');
   $B487 : WriteLn(#16' Get string');
   $E394 : WriteLn(#16' BASIC Reset');
   $E3BF : WriteLn(#16' Init BASIC Interpretor');
   $E422 : WriteLn(#16' BASIC Message');
   $E453 : WriteLn(#16' Init vectors $0300-$030B');
   $E4DA : WriteLn(#16' Set Color');
   $E518 : WriteLn(#16' Init Keyboard & Video');
   $E56C : WriteLn(#16' Set current line addr');
   $E5A0 : WriteLn(#16' Init Video');
   $E716 : WriteLn(#16' Output char to screen (device 3)');
   $E7D4 : WriteLn(#16' put MAJ on screen');
   $E9F0 : WriteLn(#16' Compute Char offset from X');
   $E9FF : WriteLn(#16' Clear line');
   $EA24 : WriteLn(#16' Compute Color offset');
   $EA31 : WriteLn(#16' Standard IRQ routine');
   $EA87 : WriteLn(#16' Check keyboard');
   $EE8E : Writeln(#16' Clock = 1');
   $EEA9 : WriteLn(#16' ');
   $F69B : WriteLn(#16' Incr. Real Time Clock');
   $FCE2 : WriteLn(#16' C=64 Entry point');
   $FCEF : WriteLn(#16' (no cartridge)');
   $FD02 : WriteLn(#16' Detect cartridge');
   $FD15 : WriteLn(#16' Restore I/O vectors');
   $FD1A : WriteLn(#16' Set I/O Vectors from XY');
   $FD50 : WriteLn(#16' Init z-page & search ROM');
   $FDA3 : WriteLn(#16' Reset I/O');
   $FDDD : WriteLn(#16' Check monitor');
   $FF5B : WriteLn(#16' Init I/O');
   $FF5E : WriteLn(#16' Wait retrace');
   $FF6E : WriteLn(#16' Set IRQs to 1/60s');
   $FF9C : WriteLn(#16' [Get/Set Bottom of memory]');
   $FF99 : WriteLn(#16' [Get/Set Top of memory]');
   $FFCF : WriteLn(#16' Start BASIC');
   $FFD2 : WriteLn(#16' [Output char]');
  end;

  str:=format('R=%3d A1=%4d/%4d B1=%4d/%4d A2=%4d/%4d B2=%4d/%4d t=%2d',
               [Vic.Raster,
                CIA1.TimerA,CIA1.Regs.TIA,
                CIA1.TimerB,CIA1.Regs.TIB,
                CIA2.TimerA,CIA2.Regs.TIA,
                CIA2.TimerB,CIA2.Regs.TIB,
                Cycles]);
  writeln(str);
  {$ifdef list}
   writeln(lst,str);
  {$endif}
  str:=format('PC=%4x OP=%2x A=%2x X=%2x Y=%2x',[PC,IR,A,X,Y]);
  str:=str+(' P=');
  str:=str+FlagStr;
  str:=str+(' S=1'+IntToHex(S,2));

  With OpCode[ir] do begin
   str:=str+(' '+Name+' ');
   Case Parm of
    pImpl : { nothing };
    pAbso : str:=str+( '$'+HexaW(pip2)+'   '+Value(pip2));
    pZPge : str:=str+( '$'+HexaB(pip1)+'     '+Value(pip1));
    pImm  : str:=str+('#$'+HexaB(pip1));
    pAbsX : str:=str+( '$'+HexaW(pip2)+'+X'+' '+Value(pip2+X));
    pAbsY : str:=str+( '$'+HexaW(pip2)+'+Y'+' '+Value(pip2+Y));
    pIndX : str:=str+('($'+HexaB(pip1)+'+X'+' '+Value(Deek(pip1+X)));
    pIndY : str:=str+('($'+HexaB(pip1)+')+Y'+' '+Value(Deek(pip1)+Y));
    pZPgX : str:=str+( '$'+HexaB(pip1)+'+X'+' '+Value(pip1+X));
    pZPgY : str:=str+( '$'+HexaB(pip1)+'+Y'+' '+Value(pip1+Y));
    pRela : str:=str+( '$'+HexaW(PC+2+ShortInt(pip1)));
    pInd  : str:=str+('($'+HexaW(pip2)+')');
   end;
  end;

//  if Dump[47]='J' then Dump[0]:=#57;

//  if (ir=$40)or irq then TextColor(LightGreen) else TextColor(Yellow);

//  WriteLn(Cycles:2,' ',Dump);
  writeln(str);
  {$IFDEF LIST}
   writeln(lst,str);
  {$ENDIF}
end;
{$ENDIF}

{$IFDEF LIST}
procedure TC64.List;
 var
  cmp:TListRecord;
 begin
  if eof(dos) then begin
   messagebox(hWnd,'6510 ok',title,0);
   halt;
  end;
  read(dos,cmp);
  if (cmp.Raster<>VIC.Raster)
  or (cmp.TimerA1<>CIA1.TimerA)
  or (cmp.TimerB1<>CIA1.TimerB)
  or (cmp.TimerA2<>CIA2.TimerA)
  or (cmp.TimerB2<>CIA2.TimerB)
  or (cmp.TIA1<>CIA1.Regs.TIA)
  or (cmp.TIB1<>CIA1.Regs.TIB)
  or (cmp.TIA2<>CIA2.Regs.TIA)
  or (cmp.TIB2<>CIA2.Regs.TIB)
  or (cmp.cycles<>cycles)
  or (cmp.pc<>pc)
  or (cmp.op<>ir)
  or (cmp.a <>a )
  or (cmp.x <>x )
  or (cmp.y <>y )
  or (cmp.s <>s )
  or (cmp.p <> byte(p)) then begin
  {$ifdef debug}
   writeln(lst,':(');
   byte(p):=cmp.p; // for FlagStr
   writeln(lst,
    'R=',cmp.Raster:3,
    ' A1=',cmp.TimerA1:4,'/',cmp.TIA1:4,
    ' B1=',cmp.TimerA2:4,'/',cmp.TIB1:4,
    ' A2=',cmp.TimerB1:4,'/',cmp.TIA2:4,
    ' B2=',cmp.TimerB2:4,'/',cmp.TIB2:4,
    ' t=',cmp.Cycles:2
   );
   writeln(lst,
    'PC=',hexaw(cmp.pc),
    ' OP=',Hexab(cmp.op),
    ' A=',Hexab(cmp.a),
    ' X=',hexab(cmp.x),
    ' Y=',hexab(cmp.y),
    ' P=',FlagStr,
    ' S=1',hexab(cmp.s)
   );
   closefile(lst);
  {$endif}
   VIC.DrawDib; // refresch screen
   messagebox(hwnd,'6510 error',title,0);
   halt;
   end;
 end;
{$ENDIF}

end.
