unit vic;

interface

uses
 windows,resource;

Const
 TopOfScreen=30;
 BottomOfScreen=270;

 EndOfScreen=312;

Type
{ VIC registers }
TVICRegs=record
case boolean of
true:(
 SprXY:array[0..15] of byte; { $D000..$D00F : Sprite X & Y }
 SprX8:byte;                 { $D010 : Sprite X - bit 8 }
 Ctrl1:byte;                 { $D011 : VIC Control Register 1 }
 _Raster:byte;               { $D012 : Read Raster }
 LightPenX, LightPenY:byte;  { $D013..$D014 : Light Pen X Y }
 SprEnable:byte;             { $D015 : Sprite Enable (1) }
 Ctrl2:byte;                 { $D016 : VIC Control Register 2 }
 Spr2Y:byte;                 { $D017 : Sprites 0-7 Expand 2x Vertical (Y) }
 VBase:byte;                 { $D018 : VIC Memory Control Register }
 IrqFlag:byte;               { $D019 : VIC Interrupt Flag Register }
 IrqMask:byte;               { $D01A : IRQ Mask Register }
 SprBack:byte;               { $D01B : Sprite to background display priority }
 SprMMC :byte;               { $D01C : Sprites 0-7 Multi-color Mode select }
 Spr2X  :byte;               { $D01D : Sprites 0-7 Expand 2x Horizontal (X) }
 SprSCD :byte;               { $D01E : Sprite to Sprite Collision Detect }
 SprBCD :byte;               { $D01F : Sprite to Background Collision Detect }
 Border :byte;               { $D020 : Border Color }
 Back0  :byte;               { $D021 : Background color 0 }
 Back1  :byte;               { $D022 : Background color 1 }
 Back2  :byte;               { $D023 : Background color 2 }
 Back3  :byte;               { $D024 : Background color 3 }
 SprMCR0:byte;               { $D025 : Sprite Multi-Color Register 0 }
 SprMCR1:byte;               { $D026 : Sprite Multi-Color Register 1 }
 SprColor:array[$27..$2E] of byte; { $D027..$D02E : Sprites color }
);
false:(
 raw:array[0..$3F] of byte;
);
end;

 { Replace the TBitmapInfo for dib_pal_colors on StretchDiBits API }
 TPalBitmapInfo=record
  bmiHeader:TBitmapInfoHeader;
  bmiColors:Array[0..15] of Word;
 end;
 TPalette=record
  VER:word; // $0300
  NBR:word; // 16
  DAC:array[0..15,0..3] of byte;
 end;
 TCHARGEN=array[0..255,0..7] of byte;

 TVIC=Class
  DIB:Array[0..239,0..359] of byte;
  DIBInfo:TPalBitmapInfo;
  PAL:HPalette;
  Color:array[$D800..$DBFF] of byte;
  Palette:TPalette;
  VideoY:integer;
  VideoX:integer;
  TextRow:integer;
 // Registers
  Regs     :TVICRegs;
  Raster   :integer;
  IrqRaster:integer;
  VICIRQ   :boolean;
  XScroll  :integer;
  YScroll  :integer;
  FirstLine:integer;
  LastLine :integer;
  DisplayEnabled:boolean;
  DisplayMode:integer;
  TextCols:integer;
  Bank:integer;
  Video,Bitmap:integer;
  CHARGEN:^TCHARGEN;
  ColorMatrix:array[1..40] of byte;
  VideoMatrix:array[1..40] of byte;
  CharsMatrix:array[1..40] of byte;
  SkipCount:integer;
  SkipFrames:integer;
  Constructor Create;
  Destructor Destroy; override;
  Procedure Reset;
  Function EmulateLine:integer;
  Procedure StandardText;
  Procedure DrawDIB;
  Procedure NewBase;
  Procedure VBlank;
  Procedure RasterIRQ;
  Procedure Poke(Addr:word; Value:byte);
  Function  Peek(Addr:word):byte;
 end;

implementation

uses c64;

Const
 clBlack      =0;
 clWhite      =1;
 clRed        =2;
 clCyan       =3;
 clPurple     =4;
 clGreen      =5;
 clBlue       =6;
 clYellow     =7;
 clOrange     =8;
 clBrown      =9;
 clLightRed   =10;
 clDarkGrey   =11;
 clGrey       =12;
 clLightGreen =13;
 clLightBlue  =14;
 clLightGrey  =15;


Constructor TVIC.Create;
 var
  i:integer;
 begin
  SkipFrames:=8;
  LoadData(Palette,'CBM64_PAL');
  PAL:=CreatePalette(pLogPalette(@Palette)^);
  with DIBInfo do begin
   with bmiheader do begin
    bisize:=40;
    biwidth:=360;
    biheight:=240;
    biplanes:=1;
    bibitcount:=8;
    bicompression:=BI_RGB;
    bisizeimage:=0;
    bixpelspermeter:=0;
    biypelspermeter:=0;
    biclrused:=16;
    biclrimportant:=0;
   end;
   for i:=0 to 15 do bmiColors[i]:=i;
  end;
 end;

Destructor TVIC.Destroy;
 begin
  DeleteObject(PAL);
  Inherited Destroy;
 end;

Procedure TVIC.Reset;
 begin
  FillChar(Regs,SizeOf(Regs),0);
  DisplayEnabled:=false;
  YScroll:=0;
  XScroll:=0;
  FirstLine:=51;
  LastLine:=250;
  Raster:=0;
 end;

Function TVIC.EmulateLine:integer;
 var
  VideoOfs:integer;
 begin
 { Default CPU left }
  Result:=64;
 { Raster count }
  Inc(Raster);
 { End of Screen ? }
  if Raster=EndOfScreen then VBlank;
 { IRQ Raster }
  if Raster=IrqRaster then with Regs do begin
   IrqFlag:=IrqFlag or 1;
   if (IrqMask and 1)=1 then begin
    IrqFlag:=IrqFlag or $80;
    VICIRQ:=True;
   end;
  end;
 { Skip Frame ? }
  if SkipCount>0 then exit;
 { Within the visible range? }
   If (Raster<TopOfScreen)or(Raster>=BottomOfScreen) then exit;

   VideoY:=239-(Raster-TopOfScreen);
   VideoX:=20;

   if (DisplayEnabled=false)or(raster<FirstLine)or(Raster>LastLine) then begin
    FillChar(DIB[VideoY],360,Regs.Border);
    exit;
   end;

   with Regs do begin

    VideoOfs:=40*((Raster-FirstLine) shr 3);

    { Bad line condition }
    if ((Raster-FirstLine) and 7)=0 {YScroll} then begin
     {$ifdef debug}
      // writeln(lst,'>Bad line !');
     {$endif}
     Result:=44;
     case DisplayMode of
      0: begin
          Move(Color[VideoOfs+$D800],ColorMatrix,40);
          Move(CBM64.RAM[Video+VideoOfs],VideoMatrix,40);
         end;
//      2: begin
//          Move(Video^[VideoOfs],ColorMatrix,40);
//         end;
     end;
     TextRow:=0;
    end;

    FillChar(DIB[VideoY],20,Border); // Left border

    { Background on the left if XScroll>0 }
    if XScroll>0 then begin
     FillChar(DIB[VideoY,VideoX],XScroll,Back0);
     inc(VideoX,XScroll);
    end;

    case DisplayMode of
     0 : StandardText;
  //   2 : StandardBitmap;
    end;

//    Inc(VideoX,TextCols shl 3);
    Inc(TextRow);

    FillChar(DIB[VideoY,VideoX],360-VideoX,Border);

  end;

 end;

Procedure TVIC.StandardText;
 var
  i,c,col,bit:integer;
  b:byte;
 begin
 // Background color
  FillChar(DIB[VideoY,VideoX],TextCols shl 3,Regs.Back0);
 // Now the Text
  for i:=1 to TextCols do begin // for each col
   c:=VideoMatrix[i]; // Char Data index
   col:=ColorMatrix[i]; // Char color
   b:=CHARGEN^[c,TextRow];
   for bit:=0 to 7 do begin
    if (b and 128)>0  then DIB[VideoY,VideoX]:=col;
    b:=(b shl 1) and $FF;
    inc(VideoX);
   end;
  end;
 end;

Procedure TVIC.DrawDIB;
 var
  DC:HDC;
 begin
  DC:=GetDC(HWnd);
  SelectPalette(DC,PAL,False);
  RealizePalette(DC);
  StretchDIBits(
  DC,	// handle of device context
  0,	// x-coordinate of upper-left corner of dest. rect.
  0,	// y-coordinate of upper-left corner of dest. rect.
  ScreenWidth,	// width of destination rectangle
  ScreenHeight,	// height of destination rectangle
  0,	// x-coordinate of upper-left corner of source rect.
  0,	// y-coordinate of upper-left corner of source rect.
  360,	// width of source rectangle
  240,	// height of source rectangle
  @DIB,	// address of bitmap bits
  pBitmapInfo(@DIBInfo)^,	// address of bitmap data
  dib_pal_colors,	// usage
  SRCCOPY 	// raster operation code
  );
  ReleaseDC(HWnd,DC);
end;

Procedure TVIC.NewBase;
 Var
  Addr:word;
 begin
 { Video }
  Video:=Bank+(Regs.VBase and $F0) shl 6;
 { Charset }
  Addr:=Bank+(Regs.VBase and $0E) shl 10;
  if (Addr and $7000)=$1000 then
   CHARGEN:=@CBM64.CHARS[(Addr and $FFF) or $D000]
  else
   CHARGEN:=@CBM64.RAM[Addr];
 { Bitmap }
  Bitmap:=Bank+(Regs.VBase and $08) shl 10;
 end;

Procedure TVIC.VBlank;
 begin
  Raster  :=0;
  TextRow :=0;
//  DrawDIB;
  CBM64.CheckKeyboard;
  if SkipCount=0 then begin
   InvalidateRect(hWnd,nil,False);
   SkipCount:=SkipFrames;
  end else begin
   Dec(SkipCount);
  end;
 end;

procedure TVIC.RasterIRQ;
 begin
 with Regs do begin
  IrqFlag:=IrqFlag or 1;
  if (IrqMask and 1)=1 then begin
   IrqFlag:=IrqFlag or $80;
   VICIRQ:=True;
  end;
 end;
 end;


Procedure TVIC.Poke(Addr:word;Value:byte);
 var
  i:word;
 begin
 with Regs do
  case Addr and $3F of
   $11 : begin
          Ctrl1:=Value;
          YScroll:=Value and 7; { smooth scrool }
          i:=(IrqRaster and $FF) or ((Value and $80) shl 1);
          if (i<>IrqRaster)and(i=Raster) then RasterIRQ;
          IrqRaster:=i;
        (*
          if (Value and 8)=0 then begin { Select 24/25 Row Text Display }
          { 24 text lines/192 pixels }
           FirstLine:=55;
           LastLine :=246;
          end else begin
          { 25 text lines/200 pixels }
           FirstLine:=51;
           LastLine :=250;
          end;
         *)
          DisplayEnabled:=(Ctrl1 and $10)=$10;

          DisplayMode:=(Ctrl1 and $60) or (Ctrl2 and $10) shr 4;
         end;
   $12 : begin
          i:=(IrqRaster and $FF00) or Value;
          if (i<>IrqRaster)and(i=Raster) then RasterIRQ;
          IrqRaster:=i;
         end;
   $16 : begin
          Ctrl2:=Value;
          XScroll:=Value and 7;
//          Border40:=Value and 8;

          if Value and 8=0 then TextCols:=38 else TextCols:=40;

          DisplayMode:=(Ctrl1 and $60) or (Ctrl2 and $10) shr 4;
         end;
   $18 : begin
          VBase:=Value;
          NewBase;
         end;
   $19 : begin
          IrqFlag:=IrqFlag and ((not value) and $0F);
          {
	  the_cpu->ClearVICIRQ();	// Clear interrupt (hack!)
	  if (irq_flag & irq_mask) // Set master bit if allowed interrupt still pending
	   irq_flag |= 0x80;
          }
         end;
   $1A : begin
          IrqMask:=Value and $0F;
          {
	  if (irq_flag & irq_mask) ( // Trigger interrupt if pending and now allowed
	   irq_flag |= 0x80;
	   the_cpu->TriggerVICIRQ();
	  ) else (
	   irq_flag &= 0x7f;
	   the_cpu->ClearVICIRQ();
	  )
          }
         end;
  else Raw[Addr and $3F]:=Value;
  end;
 end;

Function TVIC.Peek(Addr:word):byte;
 begin
 { $D000..$D300 }
 with Regs do
  case Addr and $3F of
   $11 : Result:=(Ctrl1 and $7F) or ((Raster and $100) shr 1);
   $12 : Result:=Raster and $FF;
   $16 : Result:=Ctrl2 or $C0;
   $18 : Result:=VBase or 1;
   $19 : Result:=IrqFlag or $70;
   $1A : Result:=IrqMask or $F0;
   $1E : begin Result:=SprSCD; SprSCD:=0; end;
   $1F : begin Result:=SprBCD; SprBCD:=0; end;
   $20 : Result:=Border or $F0;
   $21 : Result:=Back0 or $F0;
   $22 : Result:=Back1 or $F0;
   $23 : Result:=Back2 or $F0;
   $24 : Result:=Back3 or $F0;
   $25 : Result:=SprMCR0 or $F0;
   $26 : Result:=SprMCR1 or $F0;
   $27..$2E : Result:=SprColor[Addr and $3F] or $F0;
  else
   Result:=Raw[Addr and $3F];
  end;
 end;

end.
